import json
from kubernetes import config, client
import click
import subprocess
from packaging.version import Version
from pkg_resources import resource_filename
from tensorkube.constants import get_cluster_name, CliColors, get_image_registry_id
from tensorkube.services.aws_service import get_karpenter_namespace, get_aws_account_id, get_session_region, \
    get_availability_zones
from tensorkube.services.k8s_service import add_helm_annotations_and_labels, get_helm_release_version, \
    get_unhealthy_nodes, get_tensorkube_cluster_context_name
from tensorkube.services.configure_service import update_cfn_configure_stack
from tensorkube.services.iam_service import create_iam_policy, attach_role_policy, detach_role_policy, delete_policy
from tensorkube.services.eks_service import install_karpenter_with_command


def add_do_not_disrupt_annotation_to_keda_pods():
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        raise Exception("Unable to get cluster context")
    k8s_api_client = config.new_client_from_config(context=context_name)

    # Load the existing kubeconfig
    v1 = client.CoreV1Api(k8s_api_client)
    namespace = "keda"

    # List all pods in the keda namespace
    pods = v1.list_namespaced_pod(namespace=namespace)

    for pod in pods.items:
        # Check if the pod is in Pending or Running state
        if pod.status.phase in ["Pending", "Running"]:
            if ("keda-admission-webhooks" in pod.metadata.name
                or "keda-operator" in pod.metadata.name
                    or "keda-operator-metrics-apiserver" in pod.metadata.name):
                continue
            annotations = pod.metadata.annotations or {}
            annotations["karpenter.sh/do-not-disrupt"] = "true"

            # Patch the pod with the updated annotations
            body = {
                "metadata": {
                    "annotations": annotations
                }
            }
            v1.patch_namespaced_pod(name=pod.metadata.name, namespace=namespace, body=body)
            print(f"Added do-not-disrupt annotation to pod: {pod.metadata.name}")


def apply(test: bool = False):
    try:
        cluster_name = get_cluster_name()
        region = get_session_region()
        unhealthy_nodes = get_unhealthy_nodes(cluster_name=cluster_name)
        if unhealthy_nodes:
            click.echo(click.style("Nodes in Not Ready state found. If these are unhealthy, please manually delete them and rerun the migration."
                                   "Please contact support if you are unsure or need any other help.", fg=CliColors.ERROR.value))
            raise Exception("NotReady nodes found. If these are unhealthy, please manually delete them and rerun the migration. Please contact support if you are unsure or need any other help.")

        account_id = get_aws_account_id()
        karpenter_version = "0.37.7"
        karpenter_namespace = get_karpenter_namespace()
        karpenter_iam_role_arn = f"arn:aws:iam::{account_id}:role/{cluster_name}-karpenter"

        click.echo("Adding Helm labels and annotations to Karpenter CRDs")

        add_helm_annotations_and_labels(resource_api_version="apiextensions.k8s.io/v1",
                                        resource_type="CustomResourceDefinition",
                                        resource_name="ec2nodeclasses.karpenter.k8s.aws", release_name="karpenter-crd",
                                        release_namespace="kube-system", resource_namespace=None)
        add_helm_annotations_and_labels(resource_api_version="apiextensions.k8s.io/v1",
                                        resource_type="CustomResourceDefinition",
                                        resource_name="nodeclaims.karpenter.sh", release_name="karpenter-crd",
                                        release_namespace="kube-system", resource_namespace=None)
        add_helm_annotations_and_labels(resource_api_version="apiextensions.k8s.io/v1",
                                        resource_type="CustomResourceDefinition",
                                        resource_name="nodepools.karpenter.sh", release_name="karpenter-crd",
                                        release_namespace="kube-system", resource_namespace=None)



        installed_karpenter_version = get_helm_release_version("karpenter", "kube-system")
        installed_karpenter_crd_version = get_helm_release_version("karpenter-crd", "kube-system")

        if Version(installed_karpenter_version) < Version(karpenter_version):
            click.echo(f"Upgrading Karpenter to version {karpenter_version}")
            install_command = ["helm", "upgrade", "--install", "karpenter", "oci://public.ecr.aws/karpenter/karpenter",
                 f"--version", karpenter_version, "--namespace", f"{karpenter_namespace}", "--create-namespace",
                 "--set", f'serviceAccount.annotations.eks\\.amazonaws\\.com/role-arn={karpenter_iam_role_arn}',
                 "--set", f"settings.clusterName={cluster_name}",
                 "--set", f"settings.interruptionQueue={cluster_name}",
                 "--set", "controller.resources.requests.cpu=1",
                 "--set", "controller.resources.requests.memory=1Gi",
                 "--set", "controller.resources.limits.cpu= 1",
                 "--set", "controller.resources.limits.memory=1Gi",
                 "--set", "webhook.enabled=true",
                 "--set", "webhook.port=8443",
                  "--wait"]

            install_karpenter_with_command(install_command)
            click.echo("Upgrade Complete")
        else:
            click.echo(f"Karpenter version is already >= {karpenter_version}")

        if not installed_karpenter_crd_version or Version(installed_karpenter_crd_version) < Version(karpenter_version):
            click.echo(f"Upgrading Karpenter CRDs to version {karpenter_version}")
            install_karpenter_crd_command = ["helm", "upgrade", "--install", "karpenter-crd", "oci://public.ecr.aws/karpenter/karpenter-crd",
                  f"--version", karpenter_version, "--namespace", f"{karpenter_namespace}", "--create-namespace",
                  "--set", "webhook.enabled=true",
                  "--set", "webhook.serviceName='karpenter'",
                  "--set", "webhook.port=8443",
                  "--wait"]
            install_karpenter_with_command(install_karpenter_crd_command)
            click.echo("Upgrade Complete")
        else:
            click.echo(f"Karpenter CRDs version is already >= {karpenter_version}")

        karpenter_version_v1_0 = "1.0.9"
        file_path = "configurations/aws_configs/karpenter_controller_policy_v1.json"
        with open(resource_filename("tensorkube", file_path), "r") as f:
            policy_document = f.read()
        updated_policy_document = policy_document.replace("${AWS::Partition}", "aws")\
            .replace("${AWS::Region}", region)\
            .replace("${AWS::AccountId}", account_id)\
            .replace("${ClusterName}", cluster_name)\
            .replace("${KarpenterInterruptionQueue.Arn}", f"arn:aws:sqs:{region}:{account_id}:{cluster_name}")\
            .replace("${KarpenterNodeRole.Arn}", f"arn:aws:iam::{account_id}:role/KarpenterNodeRole-{cluster_name}")


        role_name = f"{cluster_name}-karpenter"
        policy_name = f"KarpenterControllerPolicy-{cluster_name}-v1"

        click.echo("Creating temporary IAM policy for karpenter")
        create_iam_policy(policy_name=policy_name, policy_document=json.loads(updated_policy_document))
        click.echo("Attatching temp IAM policy to role")
        attach_role_policy(account_no=account_id, policy_name=policy_name, role_name=role_name)
        click.echo("Temporary IAM polilcy created and attached to role")


        installed_karpenter_version = get_helm_release_version("karpenter", "kube-system")
        installed_karpenter_crd_version = get_helm_release_version("karpenter-crd", "kube-system")

        if Version(installed_karpenter_crd_version) < Version(karpenter_version_v1_0):
            click.echo(f"Upgrading karpenter CRDs to version {karpenter_version_v1_0}")
            install_karpenter_crd_command = ["helm", "upgrade", "--install", "karpenter-crd", "oci://public.ecr.aws/karpenter/karpenter-crd",
                 "--version", f"{karpenter_version_v1_0}", "--namespace", f"{karpenter_namespace}", "--create-namespace",
                 "--set", "webhook.enabled=true",
                 "--set", 'webhook.serviceName="karpenter"',
                 "--set", "webhook.port=8443"]
            install_karpenter_with_command(install_karpenter_crd_command)
            click.echo("Upgrade Complete")
        else:
            click.echo(f"Karpenter CRDs version is already >= {karpenter_version_v1_0}")

        if Version(installed_karpenter_version) < Version(karpenter_version_v1_0):
            click.echo(f"Upgrading karpenter to version {karpenter_version_v1_0}")
            install_karpenter_command = ["helm", "upgrade", "--install", "karpenter", "oci://public.ecr.aws/karpenter/karpenter",
                 "--version", f"{karpenter_version_v1_0}", "--namespace", f"{karpenter_namespace}", "--create-namespace",
                 "--set", f'serviceAccount.annotations.eks\\.amazonaws\\.com/role-arn={karpenter_iam_role_arn}',
                 "--set", f'settings.clusterName={cluster_name}',
                 "--set", f"settings.interruptionQueue={cluster_name}",
                 "--set", "controller.resources.requests.cpu=1",
                 "--set", "controller.resources.requests.memory=1Gi",
                 "--set", "controller.resources.limits.cpu= 1",
                 "--set", "controller.resources.limits.memory=1Gi",
                 "--wait"]
            install_karpenter_with_command(install_karpenter_command)
            click.echo("Upgrade Complete")
        else:
            click.echo(f"Karpenter version is already >= {karpenter_version_v1_0}")

        updated_params = {
            "TemplatesVersion": "v0.0.6",
            "AWSAccessLambdaFunctionImageVersion": "v1.0.2",
            "CliVersion": "0.0.89",
            "EksAccessLambdaFunctionImageVersion": "v1.0.4"
        }
        click.echo("Updating CloudFormation templates")
        update_cfn_configure_stack(updated_parameters=updated_params, test=test)
        click.echo("CloudFormation templates updated")


        click.echo("Detatching and deleting temporary IAM policy")
        detach_role_policy(account_no=account_id, role_name=role_name, policy_name=policy_name)
        delete_policy(account_no=account_id, policy_name=policy_name)
        click.echo("Temporary IAM policy deleted")

        click.echo("Setting job pods to do-not-disrupt")
        add_do_not_disrupt_annotation_to_keda_pods()

        click.echo("Uprading karpenter Nodepools and Nodeclasses to v1")
        release_name = "karpenter-configuration"
        chart_version = "0.1.3"
        update_karpenter_config_command = ["helm", "upgrade", "--install", release_name,
                        f"oci://public.ecr.aws/{get_image_registry_id(test)}/tensorfuse/helm-charts/tk-karpenter-config",
                        "--version", chart_version,
                        "--set", f"clusterName={get_cluster_name()}"]
        availability_zones = get_availability_zones(region=region)
        for (i, az) in enumerate(availability_zones):
            update_karpenter_config_command.append("--set")
            update_karpenter_config_command.append(f"karpenterTopologyZoneValues[{i}]={az}")
        subprocess.run(update_karpenter_config_command,check=True)
        click.echo("Upgrade complete")

        installed_karpenter_version = get_helm_release_version("karpenter", "kube-system")
        installed_karpenter_crd_version = get_helm_release_version("karpenter-crd", "kube-system")
        installed_karpenter_config_version = get_helm_release_version("karpenter-configuration", "default")


        if Version(installed_karpenter_config_version) < Version("0.1.1"):
            raise Exception(f"Karpenter Nodeclasses and Nodepools not up to date")
        karpenter_version_v1_4 = "1.4.0"

        if Version(installed_karpenter_crd_version) < Version(karpenter_version_v1_4):
            click.echo(f"Upgrading karpenter CRDs to version {karpenter_version_v1_4}")
            install_karpenter_crd_command = ["helm", "upgrade", "--install", "karpenter-crd", "oci://public.ecr.aws/karpenter/karpenter-crd",
                 "--version", f"{karpenter_version_v1_4}", "--namespace", f"{karpenter_namespace}", "--create-namespace",
                 "--set", "webhook.enabled=true",
                 "--set", 'webhook.serviceName="karpenter"',
                 "--set", "webhook.port=8443"
                 "--wait"]
            install_karpenter_with_command(install_karpenter_crd_command)
            click.echo("Upgrade complete")
        else:
            click.echo(f"Karpenter CRDs version is already >= {karpenter_version_v1_4}")

        if Version(installed_karpenter_version) < Version(karpenter_version_v1_4):
            click.echo(f"Upgrading karpenter to version {karpenter_version_v1_4}")
            install_karpenter_command = ["helm", "upgrade", "--install", "karpenter", "oci://public.ecr.aws/karpenter/karpenter",
                 "--version", f"{karpenter_version_v1_4}", "--namespace", f"{karpenter_namespace}", "--create-namespace",
                 "--set", f'serviceAccount.annotations.eks\\.amazonaws\\.com/role-arn={karpenter_iam_role_arn}',
                 "--set", f'settings.clusterName={cluster_name}',
                 "--set", f"settings.interruptionQueue={cluster_name}",
                 "--set", "controller.resources.requests.cpu=1",
                 "--set", "controller.resources.requests.memory=1Gi",
                 "--set", "controller.resources.limits.cpu= 1",
                 "--set", "controller.resources.limits.memory=1Gi",
                 "--set", "settings.featureGates.nodeRepair=true",
                 "--wait"]
            install_karpenter_with_command(install_karpenter_command)
            click.echo("Upgrade complete")
        else:
            click.echo(f"Karpenter version is already >= {karpenter_version_v1_4}")

        click.echo("Karpenter upgrade complete")

    except Exception as e:
        click.echo(click.style(f"Failed to upgrade karpenter: {e}", bold=True, fg="red"))
        raise e


from tensorkube.services.s3_access_service import create_s3_access_to_pods
import click

def apply(test: bool = False):
    try:
        create_s3_access_to_pods()
    except Exception as e:
        raise e



import click
from tensorkube.services.configure_service import update_cfn_configure_stack
from tensorkube.helpers import is_valid_email

def apply(test: bool = False):
    try:
        user_alert_email = click.prompt("Enter the email address for cluster issue alerts", type=str, )
        if not is_valid_email(user_alert_email):
            raise ValueError("Invalid email address provided for alerts.")
        updated_params = {
            "TemplatesVersion": "v0.0.7",
            "AWSAccessLambdaFunctionImageVersion": "v1.0.2",
            "CliVersion": "0.0.92",
            "EksAccessLambdaFunctionImageVersion": "v1.0.4",
            "MonitoringLambdaFunctionImageVersion": "v1.0.0",
            "UserAlertEmail": user_alert_email,
            "EvaluationPeriods": "45",
            "RepeatedNotificationPeriod": "900",
            "TagForRepeatedNotification": "TensorkubeRepeatedAlarm:true"
        }
        click.echo("Updating CloudFormation templates")
        update_cfn_configure_stack(updated_parameters=updated_params, test=test)
        click.echo("CloudFormation templates updated")

    except Exception as e:
        click.echo(click.style(f"Failed to update base stack: {e}", bold=True, fg="red"))
        raise e


import click
import subprocess
from tensorkube.constants import get_cluster_name, get_image_registry_id
from tensorkube.services.aws_service import get_availability_zones, get_session_region

def apply(test: bool = False):
    try:
        cluster_name = get_cluster_name()
        region = get_session_region()
        release_name = "karpenter-configuration"
        chart_version = "0.1.4"

        click.echo(click.style("Updating Karpenter configuration to add p4de support for A100 GPUs", bold=True))

        # Build helm upgrade command
        update_karpenter_config_command = [
            "helm", "upgrade", "--install", release_name,
            f"oci://public.ecr.aws/{get_image_registry_id(test)}/tensorfuse/helm-charts/tk-karpenter-config",
            "--version", chart_version,
            "--set", f"clusterName={cluster_name}"
        ]

        # Add availability zones
        availability_zones = get_availability_zones(region=region)
        for (i, az) in enumerate(availability_zones):
            update_karpenter_config_command.append("--set")
            update_karpenter_config_command.append(f"karpenterTopologyZoneValues[{i}]={az}")

        # Run helm upgrade
        subprocess.run(update_karpenter_config_command, check=True)

        click.echo(click.style("✓ Successfully added p4de support to GPU nodepools", bold=True, fg="green"))
        click.echo(click.style("  • New A100 deployments will use p4de instances (80GB A100)", fg="yellow"))
        click.echo(click.style("  • Existing deployments continue on p4d until redeployed", fg="yellow"))
        click.echo(click.style("  • Both p4d and p4de instances are now supported", fg="yellow"))

    except subprocess.CalledProcessError as e:
        click.echo(click.style(f"Failed to update Karpenter configuration: {e}", bold=True, fg="red"))
        raise e
    except Exception as e:
        click.echo(click.style(f"Unexpected error during migration: {e}", bold=True, fg="red"))
        raise e



from typing import Optional

from tensorkube.services.aws_service import get_sqs_client
from tensorkube.constants import get_cluster_name
def get_sqs_access_policy_name():
    return f"{get_cluster_name()}-sqs-access-policy"


def create_sqs_queue(queue_name: str, cluster_region: Optional[str] = None) -> str:
    sqs = get_sqs_client(region=cluster_region)
    response = sqs.create_queue(
        QueueName=queue_name,
        Attributes={
            'DelaySeconds': '0',
            'MessageRetentionPeriod': '86400',  # Retain messages for 1 day
            'VisibilityTimeout': '43200' # Make messages invisible for 12 hrs
        }
    )
    queue_url = response['QueueUrl']
    return queue_url


def delete_sqs_queue(queue_name: str) -> None:
    sqs = get_sqs_client()
    queue_url = sqs.get_queue_url(QueueName=queue_name)['QueueUrl']
    sqs.delete_queue(QueueUrl=queue_url)
    print(f"Queue {queue_name} deleted successfully.")

def queue_message(queue_url: str, message_body: str, cluster_region: Optional[str] = None) -> None:
    sqs = get_sqs_client(region=cluster_region)
    response = sqs.send_message(
        QueueUrl=queue_url,
        MessageBody=message_body
    )
    print(f"Message sent to queue")



import boto3
import os
import yaml
import time
import subprocess
from kubernetes import client, config
from typing import Dict
from enum import Enum
import json
import datetime

KUBECONFIG_FILE_PATH = '/tmp/kubeconfig'
# KUBECONFIG_FILE_PATH = '~/.kube/config'


def get_aws_caller_identity():
    return boto3.client('sts').get_caller_identity()


def generate_kubeconfig(cluster_name: str, region: str, aws_account_id: str):
    eks_client = boto3.client('eks', region_name=region)

    cluster_info = eks_client.describe_cluster(name=cluster_name)['cluster']

    # Extract cluster details
    endpoint = cluster_info['endpoint']
    cluster_ca_certificate = cluster_info['certificateAuthority']['data']

    # Create a kubeconfig dictionary
    kubeconfig_data = {
        "apiVersion": "v1",
        "clusters": [{
            "cluster": {
                "server": endpoint,
                "certificate-authority-data": cluster_ca_certificate
            },
            "name": f'arn:aws:eks:{region}:{aws_account_id}:cluster/{cluster_name}'
        }],
        "contexts": [{
            "context": {
                "cluster": f'arn:aws:eks:{region}:{aws_account_id}:cluster/{cluster_name}',
                "user": f'arn:aws:eks:{region}:{aws_account_id}:cluster/{cluster_name}'
            },
            "name": f'arn:aws:eks:{region}:{aws_account_id}:cluster/{cluster_name}'
        }],
        "current-context": f'arn:aws:eks:{region}:{aws_account_id}:cluster/{cluster_name}',
        "kind": "Config",
        "users": [{
            "name": f'arn:aws:eks:{region}:{aws_account_id}:cluster/{cluster_name}',
            "user": {
                "exec": {
                    "apiVersion": "client.authentication.k8s.io/v1beta1",
                    "command": "aws",
                    "args": [
                        "eks",
                        "get-token",
                        "--cluster-name", cluster_name
                    ],
                    "env": [
                        {"name": "AWS_DEFAULT_REGION", "value": region}
                    ]
                }
            }
        }]
    }
    # kubeconfig_dir = os.path.expanduser('~/.kube')
    # kubeconfig_file = os.path.join(kubeconfig_dir, 'config')
    kubeconfig_dir = os.path.expanduser('/tmp')
    kubeconfig_file = os.path.join(kubeconfig_dir, 'kubeconfig')

    # Create the directory if it does not exist
    if not os.path.exists(kubeconfig_dir):
        os.makedirs(kubeconfig_dir, exist_ok=True)

    # Load existing kubeconfig or create a new one if it doesn't exist
    if os.path.exists(kubeconfig_file):
        with open(kubeconfig_file, 'r') as f:
            existing_config = yaml.safe_load(f) or {}
    else:
        existing_config = {'apiVersion': 'v1', 'kind': 'Config', 'clusters': [], 'contexts': [], 'users': []}

    # Merge new kubeconfig entry into existing config
    # Function to update or add an entry in the list
    def update_or_add_entry(existing_list, new_entry, key):
        for i, entry in enumerate(existing_list):
            if entry[key] == new_entry[key]:
                existing_list[i] = new_entry  # Update the existing entry
                return
        existing_list.append(new_entry)  # Add as a new entry if not found

    # Merge new kubeconfig entry into existing config
    update_or_add_entry(existing_config['clusters'], kubeconfig_data['clusters'][0], 'name')
    update_or_add_entry(existing_config['contexts'], kubeconfig_data['contexts'][0], 'name')
    update_or_add_entry(existing_config['users'], kubeconfig_data['users'][0], 'name')
    existing_config['current-context'] = kubeconfig_data['current-context']

    # Write back to the default kubeconfig file
    with open(kubeconfig_file, 'w') as f:
        yaml.dump(existing_config, f)

    os.chmod(kubeconfig_file, 0o700)


def get_nodes(cluster_name: str):
    kubeconfig_file = os.path.expanduser(KUBECONFIG_FILE_PATH)
    config.load_kube_config(config_file=kubeconfig_file)

    # Load the existing kubeconfig
    v1 = client.CoreV1Api()
    try:
        print(f"Getting unhealthy nodes for cluster: {cluster_name}")
        nodes = v1.list_node()
        return nodes
    except Exception as e:
        print(f"Failed to get nodes: {e}")
        raise Exception(f"Failed to get nodes: {e}")


def is_unhealthy_node(node) -> bool:
    if node.status.conditions:
        for condition in node.status.conditions:
            # Check for 'Ready' condition and its status should not be 'True'. values are 'True', 'False', 'Unknown'
            if condition.type == 'Ready' and condition.status == 'Unknown':
                return True
    return False

def is_node_deleting(node) -> bool:
    if node.metadata.deletion_timestamp:
        return True
    return False


def get_unhealthy_nodes(cluster_name: str):
    kubeconfig_file = os.path.expanduser(KUBECONFIG_FILE_PATH)
    config.load_kube_config(config_file=kubeconfig_file)

    # Load the existing kubeconfig
    v1 = client.CoreV1Api()
    try:
        print(f"Getting unhealthy nodes for cluster: {cluster_name}")
        nodes = v1.list_node()
        unhealthy_nodes = []
        current_nodes = []
        for node in nodes.items:
            # about condition - https://kubernetes.io/docs/reference/node/node-status/#condition
            current_nodes.append(node.metadata.name)
            if node.status.conditions:
                for condition in node.status.conditions:
                    # Check for 'Ready' condition and its status should not be 'True'. values are 'True', 'False', 'Unknown'
                    if condition.type == 'Ready' and condition.status == 'Unknown':
                        print(f"Node {node.metadata.name} is unhealthy")
                        print(f"status: {condition.status}")
                        print(f"last heartbeat time: {condition.last_heartbeat_time}")
                        print(f"message: {condition.message}")
                        print(f"reason: {condition.reason}")
                        unhealthy_nodes.append(node.metadata.name)
        print(f"Unhealthy nodes: {unhealthy_nodes}")
        return unhealthy_nodes
    except Exception as e:
        print(f"Failed to get nodes: {e}")
        raise Exception(f"Failed to get nodes: {e}")


def delete_k8s_node(node_name: str, cluster_name: str):
    kubeconfig_file = os.path.expanduser(KUBECONFIG_FILE_PATH)
    config.load_kube_config(config_file=kubeconfig_file)

    # Load the existing kubeconfig
    v1 = client.CoreV1Api()
    try:
        print(f"Deleting node: {node_name}")
        node = v1.read_node(name=node_name)
        if node.metadata.finalizers:
            print(f"Removing finalizers from node: {node_name}")
            body = {"metadata": {"finalizers": []}}
            v1.patch_node(name=node_name, body=body)
        response = v1.delete_node(name=node_name, grace_period_seconds=0, propagation_policy='Foreground', pretty=True)
        print(f"Deletion Node Response: {response}")
        return True
    except Exception as e:
        print(f"Failed to delete node: {e}")
        return False


def delete_unhealthy_nodes_and_wait(cluster_name: str):
    while True:
        nodes = get_nodes(cluster_name)
        unhealthy_nodes = False
        for node in nodes.items:
            if is_unhealthy_node(node):
                unhealthy_nodes = True
                if not is_node_deleting(node):
                    print(f"Node {node.metadata.name} is unhealthy and not deleting, deleting it now")
                    delete_k8s_node(node.metadata.name, cluster_name)
                else:
                    print(f"Node {node.metadata.name} is unhealthy and deleting")
        if not unhealthy_nodes:
            print(f"All unhealthy nodes have been deleted")
            return
        else:
            print(f"Unhealthy nodes found, waiting for 60 seconds before checking again")
            time.sleep(60)


def handler(event, context):
    records = event['Records']
    record = records[0]

    sns_msg = record['Sns']
    message_str = sns_msg['Message']
    message = json.loads(message_str)

    account_id = message.get('AWSAccountId', None)
    if not account_id:
        raise ValueError("AWSAccountId not found in message")

    trigger = message.get('Trigger', None)
    node_name = None
    cluster_name = None
    region = None
    if trigger:
        dimensions = trigger.get('Dimensions', [])
        for dimension in dimensions:
            if dimension.get('name') == 'NodeName':
                node_name = dimension.get('value')
            if dimension.get('name') == 'ClusterName':
                cluster_name = dimension.get('value')

    region = boto3.session.Session().region_name
    print(f"Cluster: {cluster_name}")
    print(f"Region: {region}")
    if not cluster_name:
        raise ValueError("ClusterName not found in message")
    if not region:
        raise ValueError("Region is not found in message")

    print('Generating kubeconfig')
    generate_kubeconfig(cluster_name, region, account_id)
    os.environ['XDG_CONFIG_HOME'] = '/tmp/.config'
    os.environ['XDG_CACHE_HOME'] = '/tmp/.cache'
    os.environ['KUBECONFIG'] = KUBECONFIG_FILE_PATH
    print('Kubeconfig generated successfully')

    print('Getting unhealthy nodes')
    unhealthy_nodes = get_unhealthy_nodes(cluster_name)
    if not unhealthy_nodes:
        print('No unhealthy nodes found')
        return
    print(f"Unhealthy nodes: {unhealthy_nodes}")
    for node_name in unhealthy_nodes:
        deleted = delete_k8s_node(node_name, cluster_name)
        if deleted:
            print(f"Node {node_name} deleted successfully")
        else:
            print(f"Failed to delete node {node_name}")
            raise Exception(f"Failed to delete node {node_name}")
    print('Unhealthy nodes deleted successfully')
    # delete_unhealthy_nodes_and_wait(cluster_name)



import click
import subprocess
from pkg_resources import resource_filename
from tensorkube.constants import get_cluster_name
from tensorkube.services.configure_service import update_cfn_configure_stack

def apply(test: bool = False):
    try:
        updated_params = {
            "TemplatesVersion": "v0.0.4",
            "AWSAccessLambdaFunctionImageVersion": "v1.0.2",
            "CliVersion": "0.0.87",
            "EksAccessLambdaFunctionImageVersion": "v1.0.2"
        }
        update_cfn_configure_stack(updated_parameters=updated_params, test=test)

    except Exception as e:
        click.echo(click.style(f"Failed to update CloudFormation Stack: {e}", bold=True, fg="red"))
        raise e


import click
from tensorkube.services.configure_service import update_cfn_configure_stack

def apply(test=False):
    try:
        updated_params = {
            "TemplatesVersion": "v0.0.9",
            "CliVersion": "0.1.0",
            "AWSAccessLambdaFunctionImageVersion": "v1.0.2",
            "EksAccessLambdaFunctionImageVersion": "v1.0.5",
            "MonitoringLambdaFunctionImageVersion": "v1.0.0"
        }
        update_cfn_configure_stack(updated_parameters=updated_params, test=test)
    except Exception as e:
        click.echo(click.style(f"Failed to update base stack: {e}", bold=True, fg="red"))
        raise e


from tensorkube.services.istio import install_istio_on_cluster
import click

def apply(test: bool = False):
    try:
        install_istio_on_cluster()
        click.echo("Successfully increased connection idle timeout")

    except Exception as e:
        raise e



import time
import click
import boto3
from kubernetes import client

from tensorkube.constants import get_cluster_name, CliColors
from tensorkube.services.k8s_service import list_namespaced_k8s_deployments, scale_k8s_deployment_replicas, \
    describe_k8s_deployment, check_nodes_ready, add_k8s_nodegroup_affinity_to_deployment, get_nodes_with_label_selector, \
    delete_node_and_wait, get_deployment_for_pod, get_pods_on_node
from tensorkube.services.eks_service import upgrade_nodegroup_ami_version, scale_nodegroup_replicas, \
    wait_for_nodegroup_active

NODEGROUP_BASE_DEPLOYMENT_NAMESPACES = ["istio-system", "knative-serving", "keda", "kube-system"]

def upgrade_managed_nodegroup_ami_version():
    cluster_name = get_cluster_name()
    nodegroup_name = f"{get_cluster_name()}-tensorkube-nodegroup"

    latest_ami_release_version = get_latest_nodegroup_release_version(cluster_name=cluster_name, nodegroup_name=nodegroup_name)
    input_version = click.prompt(click.style(
        f"This action is irreversible and will upgrade the nodegroup AMI to release version '{latest_ami_release_version}'.\nPlease input the release version to confirm you want to proceed:",
        bold=True, fg=CliColors.ERROR.value))
    if input_version != latest_ami_release_version:
        raise click.ClickException(click.style("Input version does not match the latest AMI release version. Aborting upgrade.", fg=CliColors.ERROR.value, bold=True))
    print(click.style(f"Upgrading nodegroup AMI to release version '{latest_ami_release_version}'...", fg=CliColors.INFO.value, bold=True))

    scale_nodegroup_replicas(cluster_name=cluster_name, nodegroup_name=nodegroup_name, scale=5)
    wait_for_nodegroup_active(cluster_name=cluster_name, nodegroup_name=nodegroup_name)
    wait_for_nodegroup_nodes_to_be_ready(cluster_name=cluster_name, nodegroup_name=nodegroup_name)

    add_affinity_to_deployments()
    scale_nodegroup_deployments(scale_up=True)
    wait_for_nodegroup_deployments_to_be_ready()

    upgrade_nodegroup_ami_version(cluster_name=cluster_name, nodegroup_name=nodegroup_name)
    # upgrade_nodegroup_ami_version(cluster_name=cluster_name, nodegroup_name=nodegroup_name, version="1.32.7-20250807")
    wait_for_nodegroup_active(cluster_name=cluster_name, nodegroup_name=nodegroup_name)

    scale_nodegroup_deployments(scale_up=False)
    wait_for_nodegroup_deployments_to_be_ready()

    scale_nodegroup_replicas(cluster_name=cluster_name, nodegroup_name=nodegroup_name, scale=2)
    wait_for_nodegroup_active(cluster_name=cluster_name, nodegroup_name=nodegroup_name)
    delete_cordoned_nodegroup_nodes(cluster_name=cluster_name, nodegroup_name=nodegroup_name)



UPGRADE_NODEGROUP_SCALE_UP_COUNT = 5
KUBE_SYSTEM_DEPLOYMENTS_ORIGINAL_REPLICA_COUNT = 2
DEPLOYMENTS_ORIGINAL_REPLICA_COUNT = 2
def get_new_replica_count(scale_up: bool, namespace: str) -> int:
    if scale_up:
        return UPGRADE_NODEGROUP_SCALE_UP_COUNT
    else:
        if namespace == 'kube-system':
            return KUBE_SYSTEM_DEPLOYMENTS_ORIGINAL_REPLICA_COUNT
        else:
            return DEPLOYMENTS_ORIGINAL_REPLICA_COUNT


def scale_nodegroup_deployments(scale_up: bool):
    for ns in NODEGROUP_BASE_DEPLOYMENT_NAMESPACES:
        deployments = list_namespaced_k8s_deployments(namespace=ns)
        for deployment in deployments:
            scale_k8s_deployment_replicas(deployment_name=deployment.metadata.name, namespace=ns,
                                      new_scale=get_new_replica_count(scale_up=scale_up, namespace=ns))
            print(f"Scaled deployment {deployment.metadata.name} in namespace {ns} to {get_new_replica_count(scale_up=scale_up, namespace=ns)} replicas.")


def wait_for_deployment_to_be_ready(deployment: client.V1Deployment, timeout: int = 600):
    namespace = deployment.metadata.namespace
    start_time = time.time()

    while True:
        if time.time() - start_time > timeout:
            print(f"Timeout waiting for deployment {deployment.metadata.name} in namespace {namespace} to be ready.")
            break
        current_deployment = describe_k8s_deployment(deployment_name=deployment.metadata.name, namespace=namespace)
        desired_replicas = current_deployment.spec.replicas
        ready_replicas = current_deployment.status.ready_replicas if current_deployment.status.ready_replicas else 0
        if ready_replicas >= desired_replicas:
            print(
                f"Deployment {deployment.metadata.name} in namespace {namespace} is ready with {ready_replicas}/{desired_replicas} replicas.")
            break
        else:
            print(
                f"Waiting for deployment {deployment.metadata.name} in namespace {namespace} to be ready. Current status: {ready_replicas}/{desired_replicas} replicas.")
            time.sleep(10)  # Sleep for 10 seconds before checking again


def wait_for_nodegroup_deployments_to_be_ready(timeout: int = 600):
    for ns in NODEGROUP_BASE_DEPLOYMENT_NAMESPACES:
        deployments = list_namespaced_k8s_deployments(namespace=ns)
        for deployment in deployments:
            wait_for_deployment_to_be_ready(deployment=deployment, timeout=timeout)


def wait_for_nodegroup_nodes_to_be_ready(cluster_name: str, nodegroup_name: str):
    label_selector = f"alpha.eksctl.io/cluster-name={cluster_name},alpha.eksctl.io/nodegroup-name={nodegroup_name}"

    ready = False
    while not ready:
        ready, ready_nodes = check_nodes_ready(label_selector)
        if ready and len(ready_nodes) > 0:
            print(f"All nodes in nodegroup {nodegroup_name} are ready: {ready_nodes}")
        else:
            print(f"Waiting for nodes in nodegroup {nodegroup_name} to become ready...")
            time.sleep(10)  # Check every 10 seconds

    print(f"Nodegroup {nodegroup_name} is ready.")

# NOTE: this will cause a drift from the helm chart but we don't update the helm installation after configure
# so not implementing a more complex solution for now
def add_affinity_to_deployments():
    for namespace in NODEGROUP_BASE_DEPLOYMENT_NAMESPACES:
        deployments = list_namespaced_k8s_deployments(namespace=namespace)
        for deployment in deployments:
            print(f"Adding affinity to deployment {deployment.metadata.name} in namespace knative-serving")
            add_k8s_nodegroup_affinity_to_deployment(deployment_name=deployment.metadata.name, namespace=namespace,
                                                nodegroup_name=f"{get_cluster_name()}-tensorkube-nodegroup",
                                                cluster_name=get_cluster_name())


def delete_cordoned_nodegroup_nodes(cluster_name: str, nodegroup_name: str):
    click.echo(f"Deleting cordoned nodes in nodegroup {nodegroup_name}")

    label_selector = f"alpha.eksctl.io/cluster-name={cluster_name},alpha.eksctl.io/nodegroup-name={nodegroup_name}"

    nodes = get_nodes_with_label_selector(label_selector=label_selector)

    for node in nodes:
        if node.spec and node.spec.unschedulable:
            # get pods running on the node
            try:
                node_pods = get_pods_on_node(node_name=node.metadata.name)
            except client.ApiException as e:
                continue
            # get deployments for those pods
            # scale those deployments up by 1
            scaled_deployments = []
            for pod in node_pods:
                namespace, pod_deployment_name = get_deployment_for_pod(pod_name=pod.metadata.name, namespace=pod.metadata.namespace)
                if namespace not in NODEGROUP_BASE_DEPLOYMENT_NAMESPACES:
                    continue
                if pod_deployment_name:
                    pod_deployment = describe_k8s_deployment(deployment_name=pod_deployment_name, namespace=namespace)
                    click.echo(f"Found pod {pod.metadata.name} on node {node.metadata.name} belonging to deployment {pod_deployment.metadata.name}.")
                    click.echo(f"Scaling deployment by +1 to ensure availability during node deletion.")

                    scale_k8s_deployment_replicas(deployment_name=pod_deployment.metadata.name,
                                                  namespace=pod_deployment.metadata.namespace,
                                                  new_scale=pod_deployment.spec.replicas + 1)
                    scaled_deployments.append(pod_deployment)

                    wait_for_deployment_to_be_ready(deployment=pod_deployment)

            # delete node
            click.echo(f"Deleting node {node.metadata.name}")
            delete_node_and_wait(node_name=node.metadata.name)

            # scale deployments back down
            for deployment in scaled_deployments:
                click.echo(f"Scaling deployment {deployment.metadata.name} to original replica count.")
                scale_k8s_deployment_replicas(deployment_name=deployment.metadata.name,
                                              namespace=deployment.metadata.namespace,
                                              new_scale=get_new_replica_count(scale_up=False,
                                                                              namespace=deployment.metadata.namespace))
                wait_for_deployment_to_be_ready(deployment=deployment)


# AL2023 variants (managed nodegroups use these amiTypes)
# AL2023_MAP = {
#     "AL2023_x86_64_STANDARD": "/aws/service/eks/optimized-ami/{k8s}/amazon-linux-2023/x86_64/standard/recommended/release_version",
#     "AL2023_x86_64_NVIDIA":   "/aws/service/eks/optimized-ami/{k8s}/amazon-linux-2023/x86_64/nvidia/recommended/release_version",
#     "AL2023_ARM_64_STANDARD": "/aws/service/eks/optimized-ami/{k8s}/amazon-linux-2023/arm64/standard/recommended/release_version",
#     "AL2023_ARM_64_NVIDIA":   "/aws/service/eks/optimized-ami/{k8s}/amazon-linux-2023/arm64/nvidia/recommended/release_version",
# }
def _ssm_param_for_release_version(k8s_version: str, ami_type: str) -> str:
    if ami_type != "AL2023_x86_64_STANDARD":
        raise click.ClickException(f"Unsupported amiType for SSM release_version lookup: {ami_type}. Only AL2023_x86_64_STANDARD is supported.")
    return f"/aws/service/eks/optimized-ami/{k8s_version}/amazon-linux-2023/x86_64/standard/recommended/release_version"


def get_latest_nodegroup_release_version(cluster_name: str, nodegroup_name: str, region: str = None) -> str:
    session = boto3.Session(region_name=region) if region else boto3.Session()
    eks = session.client("eks")
    ssm = session.client("ssm")

    ng = eks.describe_nodegroup(clusterName=cluster_name, nodegroupName=nodegroup_name)["nodegroup"]
    k8s_version = ng["version"]
    ami_type = ng["amiType"]

    param_name = _ssm_param_for_release_version(k8s_version, ami_type)
    value = ssm.get_parameter(Name=param_name)["Parameter"]["Value"]
    return value



import click
import importlib.metadata
from typing import Optional, List, Dict

from tensorkube.configurations.cloudformation.configure.cluster_access_lambda_files.lambda_code import \
    get_managed_node_group_name
from tensorkube.constants import get_cluster_name, get_template_bucket_name, DEFAULT_NAMESPACE, \
    get_mount_policy_name, get_mount_driver_role_name, ADDON_NAME, get_cfn_base_stack_name, get_templates_version, \
    AWS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION, EKS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION, get_image_registry_id, \
    MONITORING_LAMBDA_FUNCTION_IMAGE_VERSION, get_tag_for_repeated_alarm_notification, \
    REPEATED_NOTIFICATION_PERIOD_SECONDS, ALARM_EVALUATION_PERIODS
from tensorkube.helpers import create_mountpoint_driver_role_with_policy
from tensorkube.migration_service.migration_manager.migration_service import set_current_cli_version_to_cluster
from tensorkube.services.aws_service import get_eks_control_plane_available_zones_for_region, \
    get_aws_account_id, get_credentials, get_session_region, get_cloudformation_client
from tensorkube.services.build import get_buildkit_service_account_name, configure_buildkit_irsa
from tensorkube.services.cloudformation_service import is_existing_tk_macro, create_cloudformation_stack, \
    stream_stack_events, cloudformation, update_generic_cloudformation_stack, get_stack_parameter_value, \
    get_stack_status_from_stack_name, CfnStackStatus
from tensorkube.services.eks_service import install_karpenter, apply_knative_crds, apply_knative_core, \
    apply_nvidia_plugin, create_eks_addon
from tensorkube.services.eksctl_service import create_base_tensorkube_cluster_eksctl
from tensorkube.services.filesystem_service import configure_efs
from tensorkube.services.iam_service import create_mountpoint_iam_policy
from tensorkube.services.istio import check_and_install_istioctl, install_istio_on_cluster, install_net_istio
from tensorkube.services.job_queue_service import create_cloud_resources_for_queued_job_support, \
    create_sa_role_rb_for_job_sidecar
from tensorkube.services.k8s_service import create_aws_secret, create_build_pv_and_pvc
from tensorkube.services.karpenter_service import apply_karpenter_configuration
from tensorkube.services.knative_service import enable_knative_selectors_pv_pvc_capabilities
from tensorkube.services.local_service import check_and_install_cli_tools
from tensorkube.services.logging_service import configure_cloudwatch
from tensorkube.services.nydus import get_nydus_snapshoter_namespace, get_nydus_snapshoter_service_account_name, \
    install_nydus
from tensorkube.services.s3_access_service import create_s3_access_to_pods
from tensorkube.services.s3_service import create_s3_bucket, get_bucket_name


def cfn_configure(test: bool, user_alert_email: str = None):
    click.echo("Configuring with cloudformation...")
    cluster_name = get_cluster_name()
    stack_name = get_cfn_base_stack_name()
    create_macro = True

    template_bucket_name = get_template_bucket_name(test)

    if test:
        click.echo("Test run. Checking if Macro exists...")
        create_macro = not is_existing_tk_macro()
        click.echo("Check complete")
        user_alert_email = "divyanshu@tensorfuse.io"

    zones = get_eks_control_plane_available_zones_for_region()
    parameters = [{"ParameterKey": "ClusterName", 'ParameterValue': cluster_name},
                  {"ParameterKey": "CliVersion", 'ParameterValue': importlib.metadata.version("tensorkube")},
                  {"ParameterKey": "KedaTrainBucketName",
                   'ParameterValue': get_bucket_name(env_name='keda', type="train")},
                  {"ParameterKey": "DefaultEnv", 'ParameterValue': DEFAULT_NAMESPACE},
                  {"ParameterKey": "DefaultEnvBuildBucketName", 'ParameterValue': get_bucket_name()},
                  {"ParameterKey": "DefaultEnvBuildkitISRAServiceAccountName",
                   "ParameterValue": get_buildkit_service_account_name(namespace=DEFAULT_NAMESPACE)},
                  {"ParameterKey": "NydusNamespace", 'ParameterValue': get_nydus_snapshoter_namespace()},
                  {"ParameterKey": "NydusServiceAccountName",
                   'ParameterValue': get_nydus_snapshoter_service_account_name()},
                  {"ParameterKey": "CreateMacro", 'ParameterValue': str(create_macro)},
                  {"ParameterKey": "TemplateBucketName", "ParameterValue": template_bucket_name},
                  {"ParameterKey": "TemplatesVersion", "ParameterValue": get_templates_version()},
                  {"ParameterKey": "AWSAccessLambdaFunctionImageVersion", "ParameterValue": AWS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION},
                  {"ParameterKey": "EksAccessLambdaFunctionImageVersion", "ParameterValue": EKS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION},
                  {"ParameterKey": "MonitoringLambdaFunctionImageVersion", "ParameterValue": MONITORING_LAMBDA_FUNCTION_IMAGE_VERSION},
                  {"ParameterKey": "ImageRegistryId", "ParameterValue": get_image_registry_id(test)},
                  {"ParameterKey": "UserAlertEmail", "ParameterValue": user_alert_email},
                  {"ParameterKey": "EvaluationPeriods", "ParameterValue": ALARM_EVALUATION_PERIODS},
                  {"ParameterKey": "RepeatedNotificationPeriod", "ParameterValue": REPEATED_NOTIFICATION_PERIOD_SECONDS},
                  {"ParameterKey": "TagForRepeatedNotification", "ParameterValue": get_tag_for_repeated_alarm_notification()},
                  {"ParameterKey": "KedaBuildBucketName", "ParameterValue": get_bucket_name(env_name='keda')},
                  {"ParameterKey": "KedaBuildkitISRAServiceAccountName",
                   "ParameterValue": get_buildkit_service_account_name(namespace='keda')},
                   {"ParameterKey": "Region","ParameterValue": get_session_region()},
                  {"ParameterKey": "ManagedNodeGroupName", "ParameterValue": get_managed_node_group_name(get_cluster_name())},
                  ]
    for zone in zones:
        parameters.append({"ParameterKey": ("zone"+zone[-1]).upper(), 'ParameterValue': "true"})

    capabilities = ["CAPABILITY_NAMED_IAM", "CAPABILITY_AUTO_EXPAND"]
    creation_queued, in_created_state = create_cloudformation_stack(
        template_file_path="configurations/cloudformation/configure/tensorkube-base-stack.yaml",
        stack_name=stack_name,
        parameters=parameters,
        capabilities=capabilities)
    if creation_queued:
        stream_stack_events(stack_name)


def legacy_configure(vpc_public_subnets: str, vpc_private_subnets: str):
    if not check_and_install_cli_tools():
        return False
    # TODO!: add helm annotations

    # create cloudformation stack
    cloudformation()

    # create eks cluster
    vpc_public_subnets_list = vpc_public_subnets.split(",") if vpc_public_subnets else []
    vpc_private_subnets_list = vpc_private_subnets.split(",") if vpc_private_subnets else []
    create_base_tensorkube_cluster_eksctl(cluster_name=get_cluster_name(), vpc_public_subnets=vpc_public_subnets_list,
                                          vpc_private_subnets=vpc_private_subnets_list)
    # install karpenter
    install_karpenter()
    # # apply karpenter configuration
    apply_karpenter_configuration()
    configure_cloudwatch()
    #
    # install istio networking plane
    check_and_install_istioctl()
    install_istio_on_cluster()

    # install knative crds
    apply_knative_crds()
    # install knative core
    apply_knative_core()

    # install nvidia plugin
    apply_nvidia_plugin()
    #
    # install net istio
    install_net_istio()

    # create s3 bucket for build
    bucket_name = get_bucket_name()
    create_s3_bucket(bucket_name)

    # create mountpoint policy to mount bucket to eks cluster
    create_mountpoint_iam_policy(get_mount_policy_name(get_cluster_name()), bucket_name)

    # create s3 csi driver role and attach mountpoint policy to it
    create_mountpoint_driver_role_with_policy(cluster_name=get_cluster_name(), account_no=get_aws_account_id(),
                                              role_name=get_mount_driver_role_name(get_cluster_name()),
                                              policy_name=get_mount_policy_name(get_cluster_name()))

    # create eks addon to mount s3 bucket to eks cluster
    create_eks_addon(get_cluster_name(), ADDON_NAME, get_aws_account_id(),
                     get_mount_driver_role_name(get_cluster_name()))

    # create aws credentials cluster secret
    # TODO!: figure out how to update credentials in case of token expiry
    create_aws_secret(get_credentials())

    # create pv and pvc claims for build
    create_build_pv_and_pvc(bucket_name)

    # update knative to use pod labels
    enable_knative_selectors_pv_pvc_capabilities()

    # enable Network files system for the cluster
    click.echo("Configuring EFS for the cluster...")
    configure_efs()

    create_s3_access_to_pods()

    # install keda, create related resources
    create_cloud_resources_for_queued_job_support()
    create_sa_role_rb_for_job_sidecar()

    # configure buildkit irsa, so that pods can access ecr
    configure_buildkit_irsa()

    install_nydus()

    # set current cli version to the cluster
    set_current_cli_version_to_cluster()

    return True


def get_configured_user_alert_email():
    email = get_stack_parameter_value(get_cfn_base_stack_name(), "UserAlertEmail")
    return email

def get_parameter_value(parameter_key: str, updated_parameters: dict, test: bool = False) -> Optional[str]:
    if parameter_key == "ClusterName":
        cluster_name = get_cluster_name()
        return updated_parameters.get("ClusterName", cluster_name)
    elif parameter_key == "CliVersion":
        return updated_parameters.get("CliVersion", importlib.metadata.version("tensorkube"))
    elif parameter_key == "KedaTrainBucketName":
        return updated_parameters.get("KedaTrainBucketName", get_bucket_name(env_name='keda', type="train"))
    elif parameter_key == "DefaultEnv":
        return updated_parameters.get("DefaultEnv", DEFAULT_NAMESPACE)
    elif parameter_key == "DefaultEnvBuildBucketName":
        return updated_parameters.get("DefaultEnvBuildBucketName", get_bucket_name())
    elif parameter_key == "DefaultEnvBuildkitISRAServiceAccountName":
        return updated_parameters.get("DefaultEnvBuildkitISRAServiceAccountName",
                                               get_buildkit_service_account_name(namespace=DEFAULT_NAMESPACE))
    elif parameter_key == "NydusNamespace":
        return updated_parameters.get("NydusNamespace", get_nydus_snapshoter_namespace())
    elif parameter_key == "NydusServiceAccountName":
        return updated_parameters.get("NydusServiceAccountName", get_nydus_snapshoter_service_account_name())
    elif parameter_key == "CreateMacro":
        create_macro = True
        if test:
            click.echo("Test run. Checking if Macro exists...")
            create_macro = not is_existing_tk_macro()
            click.echo("Check complete")
        return str(create_macro)
    elif parameter_key == "TemplateBucketName":
        template_bucket_name = get_template_bucket_name(test)
        return updated_parameters.get("TemplateBucketName", template_bucket_name)
    elif parameter_key == "TemplatesVersion":
        return updated_parameters.get("TemplatesVersion", get_templates_version())
    elif parameter_key == "AWSAccessLambdaFunctionImageVersion":
        return updated_parameters.get("AWSAccessLambdaFunctionImageVersion",
                                               AWS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION)
    elif parameter_key == "EksAccessLambdaFunctionImageVersion":
        return updated_parameters.get("EksAccessLambdaFunctionImageVersion",
                                               EKS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION)
    elif parameter_key == "ImageRegistryId":
        return updated_parameters.get("ImageRegistryId", get_image_registry_id(test))
    elif parameter_key == "MonitoringLambdaFunctionImageVersion":
        return updated_parameters.get("MonitoringLambdaFunctionImageVersion",
                                      MONITORING_LAMBDA_FUNCTION_IMAGE_VERSION)
    elif parameter_key == "UserAlertEmail":
        email = updated_parameters.get("UserAlertEmail", get_configured_user_alert_email())
        if not email:
            raise click.ClickException("User Alert Email is not set")
        return email
    elif parameter_key == "EvaluationPeriods":
        return updated_parameters.get("EvaluationPeriods", ALARM_EVALUATION_PERIODS)
    elif parameter_key == "RepeatedNotificationPeriod":
        return updated_parameters.get("RepeatedNotificationPeriod", REPEATED_NOTIFICATION_PERIOD_SECONDS)
    elif parameter_key == "TagForRepeatedNotification":
        return updated_parameters.get("TagForRepeatedNotification", get_tag_for_repeated_alarm_notification())
    elif parameter_key == "KedaBuildBucketName":
        return updated_parameters.get("KedaBuildBucketName", get_bucket_name(env_name='keda'))
    elif parameter_key == "KedaBuildkitISRAServiceAccountName":
        return updated_parameters.get("KedaBuildkitISRAServiceAccountName",
                                               get_buildkit_service_account_name(namespace='keda'))
    elif parameter_key == "ManagedNodeGroupName":
        return updated_parameters.get("ManagedNodeGroupName", get_managed_node_group_name(cluster_name=get_cluster_name()))
    elif parameter_key == "Region":
        return get_session_region()
    elif parameter_key in ["ZONEA", "ZONEB", "ZONEC", "ZONED", "ZONEE", "ZONEF"]:
        return None
    else:
        raise Exception(f"Unknown parameter key : {parameter_key}")



def generate_cfn_parameters_dict_list(template_url: str, updated_parameters: dict, test: bool = False) -> List[Dict]:
    cfn_client = get_cloudformation_client()

    # Get parameters defined in the template
    response = cfn_client.get_template_summary(TemplateURL=template_url)
    parameters = response.get('Parameters', [])

    # Build Parameters list for stack operations
    parameter_list = []
    for param in parameters:
        key = param['ParameterKey']
        value = get_parameter_value(key, updated_parameters, test)
        if value is not None:
            parameter_list.append({
                'ParameterKey': key,
                'ParameterValue': value
            })

    return parameter_list

def update_cfn_configure_stack(updated_parameters, test=False):
    click.echo("Updating base cloudformation stack...")
    stack_name = get_cfn_base_stack_name()

    template_bucket_name = get_template_bucket_name(test)
    templates_version = updated_parameters.get("TemplatesVersion", get_templates_version())
    base_stack_url = f"https://{template_bucket_name}.s3.us-east-1.amazonaws.com/{templates_version}/tensorkube-base-stack.yaml"

    zones = get_eks_control_plane_available_zones_for_region()
    parameters = generate_cfn_parameters_dict_list(base_stack_url, updated_parameters, test)
    for zone in zones:
        parameters.append({"ParameterKey": ("zone" + zone[-1]).upper(), 'ParameterValue': "true"})

    capabilities = ["CAPABILITY_NAMED_IAM", "CAPABILITY_AUTO_EXPAND"]
    update_generic_cloudformation_stack(stack_name=stack_name, parameters=parameters, capabilities=capabilities,
                                        template_url = f"https://{template_bucket_name}.s3.us-east-1.amazonaws.com/{templates_version}/tensorkube-base-stack.yaml")
    stack_status = get_stack_status_from_stack_name(stack_name)
    if stack_status == CfnStackStatus.UPDATE_COMPLETE:
        click.echo(click.style("Successfully updated base stack", bold=True, fg="green"))
    else:
        click.echo(click.style(f"Failed to update base stack. Current status: {stack_status}", bold=True, fg="red"))
        raise click.ClickException(f"Failed to update base stack. Current status: {stack_status}")



import os
import platform
import shutil
import subprocess
import uuid
from typing import Optional, List

import boto3
import botocore
import click
import semver

from tensorkube.constants import CliColors
from tensorkube.constants import get_cluster_name, DEFAULT_NAMESPACE, LOCKED_AWS_CLI_VERSION
from tensorkube.services.error import CLIVersionError
import configparser

TENSORFUSE_SESSION = 'tensorfuse'

def get_eks_client(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.client("eks")

def get_session_region():
    session  = get_boto3_session()
    return session.region_name

def get_cloudformation_client(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.client("cloudformation")


def get_ec2_client(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.client("ec2")


def get_iam_client(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.client("iam")

def get_sts_client(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.client("sts")

def get_s3_client(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.client("s3")

def get_s3_resource(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.resource("s3")

def get_ecr_client(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.client("ecr")

def get_efs_client(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.client("efs")


def get_aws_account_id(session_region: Optional[str] = None):
    sts = get_sts_client(region=session_region)
    identity = sts.get_caller_identity()
    return identity['Account']

def get_logs_client(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.client("logs")

def get_sqs_client(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.client("sqs")

def get_dynamodb_client(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.client("dynamodb")

def get_dynamodb_resource(region: Optional[str] = None):
    session = get_boto3_session(region_name=region)
    return session.resource("dynamodb")

def get_aws_user_arn() -> str:
    session = get_boto3_session()
    sts = session.client('sts')
    identity = sts.get_caller_identity()
    return identity['Arn']

def get_boto3_session(region_name: Optional[str] = None):
    default_session = boto3.Session(region_name=region_name)
    config_path = os.path.expanduser(
        os.environ.get("AWS_CONFIG_FILE", "~/.aws/config")
    )

    if not default_session.available_profiles and os.environ.get("AWS_DEFAULT_REGION") is None and region_name is None:
        raise click.ClickException("No AWS profiles found. Please configure your AWS profile using the command: `aws configure sso`")
        
    if TENSORFUSE_SESSION in default_session.available_profiles:
        if not os.path.isfile(config_path):
            text =  f"AWS config file not found at {config_path}; using default session"
            click.echo(click.style(text, bold=True, fg=CliColors.WARNING.value))
            return default_session

        # load and parse
        cfg = configparser.ConfigParser()
        try:
            cfg.read(config_path)
        except Exception as e:
            text = f"Failed to read AWS config file ({e}); using default session"
            click.echo(click.style(text, bold=True, fg=CliColors.WARNING.value))
            return default_session

        profile_section = f"profile {TENSORFUSE_SESSION}"
        try:
            region = cfg.get(profile_section, "region")
            if region_name is not None:
                region = region_name
        except (configparser.NoSectionError, configparser.NoOptionError) as e:
            text = f"Profile or region not found in config ({e}); using default session"
            click.echo(click.style(text, bold=True, fg=CliColors.WARNING.value))
            return default_session

        # if we got a region, build a session with it
        if region:
            return boto3.Session(profile_name=TENSORFUSE_SESSION, region_name=region)
    return default_session

def get_session_profile_name(session_region: Optional[str] = None):
    session = get_boto3_session(region_name=session_region)
    return session.profile_name

def get_principal_arn_from_identity_center_arn(arn: str) -> str:
    arn_parts = arn.split(':')
    account_no = arn_parts[4]
    role_name = arn_parts[5].split('/')[1]
    principal_arn = f"arn:aws:iam::{account_no}:role/aws-reserved/sso.amazonaws.com/{role_name}"
    return principal_arn


def get_aws_user_principal_arn() -> str:
    sts = get_sts_client()
    identity = sts.get_caller_identity()
    if 'assumed-role' in identity['Arn']:
        return get_principal_arn_from_identity_center_arn(identity['Arn'])
    return identity['Arn']


def get_karpenter_namespace():
    return "kube-system"


def get_karpenter_version():
    return "0.37.0"


def get_credentials():
    session =  get_boto3_session()
    return session.get_credentials().get_frozen_credentials()


def are_credentials_valid(credentials):
    session = get_boto3_session()
    sts = session.client('sts', aws_access_key_id=credentials.access_key, aws_secret_access_key=credentials.secret_key,
                       aws_session_token=credentials.token)
    try:
        sts.get_caller_identity()
        return True
    except botocore.exceptions.ClientError as e:
        return False



def check_and_install_aws_cli():
    """Check if aws cli is installed and if not install it."""
    try:
        result = subprocess.run(["aws", "--version"], capture_output=True, text=True, check=True)

        # The version is in the format: "aws-cli/2.13.4 Python/3.9.6 Linux/x86_64"
        version_output = result.stdout.strip()
        # Extract the AWS CLI version using string manipulation
        aws_version = semver.VersionInfo.parse(version_output.split(" ")[0].split("/")[1])
        locked_cli_version = semver.VersionInfo.parse(LOCKED_AWS_CLI_VERSION)
        if (aws_version.major, aws_version.minor) < (locked_cli_version.major, locked_cli_version.minor):
            text = f"AWS CLI version is {aws_version}. Please upgrade AWS CLI to version above {LOCKED_AWS_CLI_VERSION}."
            click.echo(click.style(text, bold=True, fg=CliColors.ERROR.value))
            raise CLIVersionError(text)
    except Exception as e:
        if isinstance(e, CLIVersionError):
            raise e
        click.echo(
            click.style("aws-cli not found. Proceeding with installation. Might require sudo password.", bold=True,
                        fg=CliColors.WARNING.value))
        try:
            system = platform.system().lower()
            architecture = platform.machine()
            print(f"Detected operating system: {system.capitalize()}")
            print(f"Detected architecture: {architecture}")
            url = ""
            file_name = ""
            # Define the download URL and file name based on the OS
            if system == "linux":
                url = f"https://awscli.amazonaws.com/awscli-exe-linux-{architecture}-{LOCKED_AWS_CLI_VERSION}.zip"
                file_name = f"awscliv2-{LOCKED_AWS_CLI_VERSION}.zip"
            elif system == "darwin":
                url = f"https://awscli.amazonaws.com/AWSCLIV2-{LOCKED_AWS_CLI_VERSION}.pkg"
                file_name = f"AWSCLIV2-{LOCKED_AWS_CLI_VERSION}.pkg"
            else:
                print("Unsupported operating system. Please install AWS CLI manually.")
                raise Exception("Unsupported operating system.")
            # Download the installer
            print(f"Downloading AWS CLI installer from {url}...")
            subprocess.run(["curl", "-fsSL", url, "-o", file_name], check=True)
            print("AWS CLI installer downloaded successfully.")
            if system == "linux":
                # Unzip the installer
                print("Unzipping the AWS CLI installer...")
                subprocess.run(["unzip", "-o", file_name], check=True)
                print("AWS CLI installer unzipped successfully.")

                # Install the AWS CLI
                print("Installing AWS CLI...")
                subprocess.run(["sudo", "./aws/install"], check=True)
                print("AWS CLI installed successfully.")

            elif system == "darwin":
                # Use the macOS installer
                print("Installing AWS CLI using macOS package...")
                subprocess.run(["sudo", "installer", "-pkg", file_name, "-target", "/"], check=True)
                print("AWS CLI installed successfully.")
        except subprocess.CalledProcessError as e:
            print(f"Error occurred during installation: {e}")
        finally:
            if os.path.exists(file_name):
                os.remove(file_name)
            if os.path.exists("aws"):
                shutil.rmtree("aws")
            print("Clean-up completed.")

# https://github.com/eksctl-io/eksctl/blob/f4d4076ee22c9026a11a70434b802a7e0f61bd46/pkg/az/az.go#L24
# taken from eksctl source code
zoneIDsToAvoid = {
    "us-east-1": ["use1-az3"],
    "us-west-1": ["usw1-az2"],
    "ca-central-1": ["cac1-az3"],
    "cn-north-1": ["cnn1-az4"],
}


def get_eks_control_plane_available_zones_for_region():
    ec2 = get_ec2_client()
    region = get_session_region()
    response = ec2.describe_availability_zones()
    return [zone['ZoneName'] for zone in response['AvailabilityZones'] if zone['ZoneId'] not in zoneIDsToAvoid.get(region, []) and zone['ZoneType'] == "availability-zone"]


def get_availability_zones(region: str) -> List[str]:
    ec2_client = boto3.client('ec2', region_name=region)
    azs = ec2_client.describe_availability_zones(
        Filters=[{"Name": "region-name", "Values": [region]}, {"Name": "zone-type", "Values": ["availability-zone"]}],
        AllAvailabilityZones=False,
    )["AvailabilityZones"]

    zone_names = [az["ZoneName"] for az in azs if az["State"] == "available"]
    return zone_names



import click
import subprocess
from pkg_resources import resource_filename
from tensorkube.constants import get_cluster_name, get_image_registry_id
from tensorkube.services.configure_service import update_cfn_configure_stack

def apply(test: bool = False):
    try:
        release_name = "karpenter-configuration"
        chart_version = "0.1.2"
        subprocess.run(["helm", "upgrade", "--install", release_name,
                        f"oci://public.ecr.aws/{get_image_registry_id(test)}/tensorfuse/helm-charts/tk-karpenter-config",
                        "--version", chart_version,
                        "--set", f"clusterName={get_cluster_name()}"
                        ],
                       check=True)
        updated_params = {
            "TemplatesVersion": "v0.0.3",
            "AWSAccessLambdaFunctionImageVersion": "v1.0.2",
            "CliVersion": "0.0.83",
            "EksAccessLambdaFunctionImageVersion": "v1.0.2"
        }
        update_cfn_configure_stack(updated_parameters=updated_params, test=test)

    except Exception as e:
        click.echo(click.style(f"Failed to apply karpenter configuration helm chart: {e}", bold=True, fg="red"))
        raise e